{% autoescape off %}
To initiate the password reset process for your {{ user.get_username }} {{ site_name }} account,
click the link below:

{{ protocol }}://{{ domain }}{% url 'password_reset_confirm' uidb64=uid token=token %}

If clicking the link above doesn't work, please copy and paste the URL in a new browser
window instead.

If you did not request a password reset, please disregard this notice. Whoever requested it
cannot follow through on resetting your password without access to this message.

Sincerely,
{{ site_name }} Management.
{% endautoescape %}