# -*- coding: utf-8 -*-
"""
This sub-package defines the typeclass-system, a way to wrap database
access into almost-normal Python classes. Using typeclasses one can
work in normal Python while having the luxury of persistent data
storage at every turn. ObjectDB, ChannelDB, AccountDB and ScriptDB all
inherit from the models in this package. Here is also were the
Attribute and Tag models are defined along with their handlers.

"""
