{% extends "base.html" %}

{% block titleblock %}
Password Changed
{% endblock %}

{% block body %}

{% load addclass %}
<div class="container main-content mt-4" id="main-copy">
  <div class="row">
    <div class="col-lg-5 offset-lg-3 col-sm-12">
      <div class="card mt-3">
        <div class="card-body">
          <h1 class="card-title">Password Changed</h1>
          <hr />
          
          <p>Your password was changed.</p>
          
          <p>Click <a href="{% url 'index' %}">here</a> to return to the index.</p>
          
        </div>
      </div>
    </div>
  </div>
</div>
{% endblock %}
