# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-06-17 20:17


from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("comms", "0011_auto_20170606_1731"), ("comms", "0011_auto_20170217_2039")]

    operations = []
