# -*- coding: utf-8 -*-
"""
This sub-package contains Evennia's command system. It handles
everything related to parsing input from the account, building cmdsets
and executing the code associated with a found command class.

commands.default contains all the default "mux-like" commands of
Evennia.

"""
