# Generated by Django 2.2.16 on 2021-05-14 22:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("comms", "0019_auto_20210514_2032"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="msg",
            name="db_hide_from_channels",
        ),
        migrations.RemoveField(
            model_name="msg",
            name="db_receivers_channels",
        ),
        migrations.AddField(
            model_name="msg",
            name="db_receiver_external",
            field=models.CharField(
                blank=True,
                db_index=True,
                help_text="identifier for single external receiver, for use with receivers without a database existence.",
                max_length=1024,
                null=True,
                verbose_name="external receiver",
            ),
        ),
    ]
