{% extends "base.html" %}

{% block titleblock %}
Forgot Password - Reset Successful
{% endblock %}

{% block body %}

{% load addclass %}
<div class="container main-content mt-4" id="main-copy">
  <div class="row">
    <div class="col-lg-5 offset-lg-3 col-sm-12">
      <div class="card mt-3">
        <div class="card-body">
          <h1 class="card-title">Password Reset</h1>
          <hr />
          {% if user.is_authenticated %}
          <div class="alert alert-info" role="alert">You are already logged in!</div>
          {% else %}
          
          <p>Your password has been successfully reset!</p>
          
          <p>You may now log in using it <a href="{% url 'login' %}">here.</a></p>
          
          {% endif %}
        </div>
      </div>
    </div>
  </div>
</div>
{% endblock %}
