# Generated by Django 3.2.3 on 2021-05-20 21:37

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("typeclasses", "0014_alter_tag_db_category"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="attribute",
            options={"verbose_name": "Attribute"},
        ),
    ]
