# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-09-25 17:35


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("comms", "0015_auto_20170706_2041")]

    operations = [
        migrations.RemoveField(model_name="channeldb", name="db_subscriptions"),
        migrations.AlterField(
            model_name="msg", name="db_message", field=models.TextField(verbose_name="message")
        ),
    ]
