{% extends "website/base.html" %}

{% block titleblock %}To Be Implemented{% endblock %}

{% block header_ext %}
<META NAME="ROBOTS" CONTENT="NOINDEX, NOFOLLOW">
{% endblock %}

{% block content %}
<hr />
<div class="row">
    <div class="col">
        <div class="card text-center">
            <h1 class="card-header">To Be Implemented...</h1>
            <div class="card-body">
                This feature has yet to be implemented, but rest assured that it will be!
                  <br />
                  <br />
                  <br />
                  Eventually...
            </div>
        </div>
    </div>
</div>
{% endblock %}
