{% extends "website/base.html" %}


{% block header_ext %}
<META NAME="ROBOTS" CONTENT="NOINDEX, NOFOLLOW">
{% endblock %}

{% block titleblock %}Logged Out{% endblock %}

{% block body %}

{% load addclass %}
<div class="container main-content mt-4" id="main-copy">
  <div class="row">
    <div class="col-lg-5 offset-lg-3 col-sm-12">
      <div class="card mt-3">
        <div class="card-body">
          <h1 class="card-title text-center">Logged Out</h1>
            <div class="row">
              <div class="col-lg-12 col-sm-12 text-center">You have been logged out.</div>
            </div>
            <div class="row">
              <div class="col-lg-12 col-sm-12 text-center">Redirecting in 2 seconds ...</div>
            </div>
        </div>
      </div>
    </div>
  </div>
</div>

  <script type="text/JavaScript">
    setTimeout("location.href = '/';",2000);
  </script>

{% endblock %}

