# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-01-28 18:20
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("help", "0002_auto_20170606_1731")]

    operations = [
        migrations.AlterField(
            model_name="helpentry",
            name="db_entrytext",
            field=models.TextField(
                blank=True, help_text="the main body of help text", verbose_name="help entry"
            ),
        ),
        migrations.AlterField(
            model_name="helpentry",
            name="db_help_category",
            field=models.CharField(
                default="General",
                help_text="organizes help entries in lists",
                max_length=255,
                verbose_name="help category",
            ),
        ),
        migrations.AlterField(
            model_name="helpentry",
            name="db_key",
            field=models.CharField(
                help_text="key to search for", max_length=255, unique=True, verbose_name="help key"
            ),
        ),
        migrations.AlterField(
            model_name="helpentry",
            name="db_lock_storage",
            field=models.TextField(
                blank=True, help_text="normally view:all().", verbose_name="locks"
            ),
        ),
        migrations.AlterField(
            model_name="helpentry",
            name="db_tags",
            field=models.ManyToManyField(
                blank=True,
                help_text="tags on this object. Tags are simple string markers to identify, group and alias objects.",
                to="typeclasses.Tag",
            ),
        ),
    ]
