{% extends "website/base.html" %}

{% block header_ext %}
{% endblock %}

{% block content %}
<div class="row">
  <div class="col">
    <div class="card">
      <div class="card-body">
        <h1 class="card-title">{{flatpage.title}}</h1>
        {{flatpage.content}}
      </div>
    </div>
  </div>
</div>
{% endblock %}
