{% extends "website/base.html" %}

{% block titleblock %}Home{% endblock %}

{% block header_ext %}
{% endblock %}

{% block content %}
<div class="row">
    <div class="col">
        <div class="card text-center">
            <div class="card-body">

                {% include "website/homepage/main-content.html" %}

            </div>
        </div>
    </div>
</div>

<hr/>

<div class="row">
    <div class="col-12 col-md-4 mb-3">

        {% include "website/homepage/accounts-widget.html" %}

    </div>

    <div class="col-12 col-md-4 mb-3">

        {% include "website/homepage/recently-connected-widget.html" %}

    </div>

    <div class="col-12 col-md-4 mb-3">

        {% include "website/homepage/database-stats-widget.html" %}

    </div>
</div>

<div class="row">
    <div class="col">

        {% include "website/homepage/evennia-widget.html" %}

    </div>
</div>
{% endblock %}
