# Generated by Django 3.2.3 on 2021-05-20 21:37

from django.db import migrations, models

import evennia.accounts.manager


class Migration(migrations.Migration):

    dependencies = [
        ("accounts", "0009_auto_20191025_0831"),
    ]

    operations = [
        migrations.AlterField(
            model_name="accountdb",
            name="first_name",
            field=models.CharField(blank=True, max_length=150, verbose_name="first name"),
        ),
    ]
