# Generated by Django 3.2.3 on 2021-05-20 21:37

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("scripts", "0014_auto_20210520_2137"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("objects", "0011_auto_20191025_0831"),
        ("comms", "0020_auto_20210514_2210"),
    ]

    operations = [
        migrations.AlterField(
            model_name="msg",
            name="db_receiver_external",
            field=models.CharField(
                blank=True,
                db_index=True,
                help_text="Identifier for single external receiver, for use with recievers not represented by a regular database model.",
                max_length=1024,
                null=True,
                verbose_name="external receiver",
            ),
        ),
        migrations.AlterField(
            model_name="msg",
            name="db_receivers_accounts",
            field=models.ManyToManyField(
                blank=True,
                help_text="account receivers",
                related_name="receiver_account_set",
                to=settings.AUTH_USER_MODEL,
                verbose_name="Receivers (Accounts)",
            ),
        ),
        migrations.AlterField(
            model_name="msg",
            name="db_receivers_objects",
            field=models.ManyToManyField(
                blank=True,
                help_text="object receivers",
                related_name="receiver_object_set",
                to="objects.ObjectDB",
                verbose_name="Receivers (Objects)",
            ),
        ),
        migrations.AlterField(
            model_name="msg",
            name="db_receivers_scripts",
            field=models.ManyToManyField(
                blank=True,
                help_text="script_receivers",
                related_name="receiver_script_set",
                to="scripts.ScriptDB",
                verbose_name="Receivers (Scripts)",
            ),
        ),
        migrations.AlterField(
            model_name="msg",
            name="db_sender_accounts",
            field=models.ManyToManyField(
                blank=True,
                db_index=True,
                related_name="sender_account_set",
                to=settings.AUTH_USER_MODEL,
                verbose_name="Senders (Accounts)",
            ),
        ),
        migrations.AlterField(
            model_name="msg",
            name="db_sender_external",
            field=models.CharField(
                blank=True,
                db_index=True,
                help_text="Identifier for single external sender, for use with senders not represented by a regular database model.",
                max_length=255,
                null=True,
                verbose_name="external sender",
            ),
        ),
        migrations.AlterField(
            model_name="msg",
            name="db_sender_objects",
            field=models.ManyToManyField(
                blank=True,
                db_index=True,
                related_name="sender_object_set",
                to="objects.ObjectDB",
                verbose_name="Senders (Objects)",
            ),
        ),
        migrations.AlterField(
            model_name="msg",
            name="db_sender_scripts",
            field=models.ManyToManyField(
                blank=True,
                db_index=True,
                related_name="sender_script_set",
                to="scripts.ScriptDB",
                verbose_name="Senders (Scripts)",
            ),
        ),
    ]
