{% extends "base.html" %}

{% block titleblock %}
Form
{% endblock %}

{% block body %}

{% load addclass %}
<div class="container main-content mt-4" id="main-copy">
  <div class="row">
    <div class="col-lg-5 offset-lg-3 col-sm-12">
      <div class="card mt-3">
        <div class="card-body">
          <h1 class="card-title">Form</h1>
          <hr />

            {% include 'website/messages.html' %} 

            {% if form.errors %}
              {% for field in form %}
                {% for error in field.errors %}
                  <div class="alert alert-danger" role="alert">{{ error }}</div>
                {% endfor %}
              {% endfor %}
            {% endif %}
          
          <form method="post" action="?">
            {% csrf_token %}
  
            {% for field in form %}
            <div class="form-field my-3">
                {{ field.label_tag }}
                {{ field | addclass:"form-control" }}
                {% if field.help_text %}
                <small class="form-text text-muted">{{ field.help_text|safe }}</small>
                {% endif %}
            </div>
            {% endfor %}
            
            <hr />
            <div class="form-group">
              <input class="form-control btn btn-outline-secondary" type="submit" value="Submit" />
              <input type="hidden" name="next" value="{{ next }}" />
            </div>
          </form>
          
        </div>
      </div>
    </div>
  </div>
</div>
{% endblock %}
