{% extends "base.html" %}

{% block titleblock %}
{{ view.page_title }}
{% endblock %}

{% block content %}

{% load addclass %}
<div class="row">
  <div class="col">
    <div class="card">
      <div class="card-body">
        <h1 class="card-title">{{ view.page_title }}</h1>
        <hr />

          <ul>
              {% for object in object_list %}
              <li><a href="{{ object.web_get_detail_url }}">{{ object }}</a></li>
              {% endfor %}
          </ul>
          
      </div>
    </div>
  </div>
</div>
{% endblock %}
