"""
test_prototypes

This is meant to be used with Evennia unittest framework
to provide some named prototypes for use with various tests
(for example commands that accept named prototypes, not just
prototype dictionaries)
"""

GOBLIN = {"key": "Goblin", "typeclass": "DefaultCharacter", "desc": "A goblin."}

BALL = {"key": "Ball", "typeclass": "DefaultObject", "desc": "A ball."}
