<div class="card text-center">
    <h4 class="card-header text-center">Evennia</h4>

    <div class="card-body">
        <p><a href="https://evennia.com">Evennia</a> is an open-source MUD/MU*-creation framework built in
            <a href="http://python.org">Python</a>, using
            <a href="http://twistedmatrix.com">Twisted</a> and
            <a href="http://djangoproject.com">Django</a>.<br>
            Create the text-based multiplayer-game of your dreams - as
            simple or as complex as you like.</p>
    </div>
</div>