{% extends "base.html" %}

{% block titleblock %}
Forgot Password - Reset Link Sent
{% endblock %}

{% block body %}

{% load addclass %}
<div class="container main-content mt-4" id="main-copy">
  <div class="row">
    <div class="col-lg-5 offset-lg-3 col-sm-12">
      <div class="card mt-3">
        <div class="card-body">
          <h1 class="card-title">Reset Sent</h1>
          <hr />
          {% if user.is_authenticated %}
          <div class="alert alert-info" role="alert">You are already logged in!</div>
          {% else %}
          
          <p>Instructions for resetting your password will be emailed to the 
          address you provided, if that address matches the one we have on file 
          for your account. You should receive them shortly.</p>
          
          <p>Please allow up to to a few hours for the email to transmit, and be 
          sure to check your spam folder if it doesn't show up in a timely manner.</p>
          
          <hr />
          
          <p><a href="{% url 'index' %}">Click here</a> to return to the main page.</p>
          
          {% endif %}
        </div>
      </div>
    </div>
  </div>
</div>
{% endblock %}
