{% extends "base.html" %}

{% block titleblock %}
{{ view.page_title }}
{% endblock %}

{% block body %}

{% load addclass %}
<div class="container main-content mt-4" id="main-copy">
  <div class="row">
    <div class="col-lg-5 offset-lg-3 col-sm-12">
      <div class="card mt-3 border border-danger">
        <div class="card-body">
          <h1 class="card-title">{{ view.page_title }}</h1>
          <hr />
          <form method="post" action="?">
            {% csrf_token %}
            <p>Are you sure you want to delete "{{ object }}"?</p>
            
            <hr />
            <div class="form-group">
              <input class="form-control btn btn-outline-danger" type="submit" value="yes" />
              <input type="hidden" name="next" value="{{ next }}" />
            </div>
          </form>
          
        </div>
      </div>
    </div>
  </div>
</div>
{% endblock %}
