"""
Wilderness contrib - titeuf87, 2017

"""

from .wilderness import WildernessExit  # noqa
from .wilderness import WildernessMapProvider  # noqa
from .wilderness import WildernessRoom  # noqa
from .wilderness import WildernessScript  # noqa
from .wilderness import create_wilderness  # noqa
from .wilderness import enter_wilderness  # noqa
from .wilderness import get_new_coordinates  # noqa
